@@ -1,5 +1,5 @@
 """ torchsummary.py """
-from typing import Any, Dict, Generator, List, Optional, Sequence, Union
+from typing import Any, Dict, Generator, List, Mapping, Optional, Sequence, Tuple, Union
 
 import torch
 import torch.nn as nn
@@ -7,21 +7,23 @@ from torch.utils.hooks import RemovableHandle
 
 from .formatting import FormattingOptions, Verbosity
 from .layer_info import LayerInfo
-from .model_statistics import CORRECTED_INPUT_SIZE_TYPE, ModelStatistics
+from .model_statistics import CORRECTED_INPUT_SIZE_TYPE, HEADER_TITLES, ModelStatistics
 
 # Some modules do the computation themselves using parameters
 # or the parameters of children. Treat these as layers.
 LAYER_MODULES = (torch.nn.MultiheadAttention,)
 INPUT_SIZE_TYPE = Sequence[Union[int, Sequence[Any], torch.Size]]
+INPUT_DATA_TYPE = Optional[Union[torch.Tensor, torch.Size, Sequence[torch.Tensor], INPUT_SIZE_TYPE]]
+DEFAULT_COLUMN_NAMES = ("output_size", "num_params")
 
 
 def summary(
     model: nn.Module,
-    input_data: Union[torch.Tensor, torch.Size, Sequence[torch.Tensor], INPUT_SIZE_TYPE],
+    input_data: INPUT_DATA_TYPE = None,
     *args: Any,
-    batch_dim: int = 0,
+    batch_dim: Optional[int] = 0,
     branching: bool = True,
-    col_names: Sequence[str] = ("output_size", "num_params"),
+    col_names: Optional[Sequence[str]] = None,
     col_width: int = 25,
     depth: int = 3,
     device: Optional[torch.device] = None,
@@ -32,7 +34,7 @@ def summary(
     """
     Summarize the given PyTorch model. Summarized information includes:
         1) Layer names,
-        2) output shape,
+        2) input/output shapes,
         3) kernel shape,
         4) # of parameters,
         5) # of operations (Mult-Adds)
@@ -45,17 +47,23 @@ def summary(
                 Example input tensor of the model (dtypes inferred from model input).
                 - OR -
                 Shape of input data as a List/Tuple/torch.Size (dtypes must match model input,
-                default is FloatTensors).
+                default is FloatTensors). Should NOT include batch size in the tuple.
+                - OR -
+                If input_data is not provided, no forward pass through the network is performed,
+                and the provided model information is limited to layer names.
 
         batch_dim (int):
                 Batch_dimension of input data. Default: 0
+                If batch_dim is None, the input data is assumed to contain the batch dimension.
+                WARNING: in a future version of torch-summary, the default will change to None.
 
         branching (bool):
                 Whether to use the branching layout for the printed output. Default: True
 
         col_names (Sequence[str]):
                 Specify which columns to show in the output. Currently supported:
-                        ('output_size', 'num_params', 'kernel_size', 'mult_adds')
+                        ("input_size", "output_size", "num_params", "kernel_size", "mult_adds")
+                If input_data is not provided, only "num_params" is used.
                 Default: ("output_size", "num_params")
 
         col_width (int):
@@ -85,15 +93,80 @@ def summary(
         ModelStatistics object
                 See torchsummary/model_statistics.py for more information.
     """
-    assert verbose in (0, 1, 2)
+    if col_names is None:
+        col_names = ("num_params",) if input_data is None else DEFAULT_COLUMN_NAMES
+
+    validate_user_params(input_data, col_names, verbose)
+    input_size = []  # type: CORRECTED_INPUT_SIZE_TYPE
     summary_list = []  # type: List[LayerInfo]
-    hooks = []  # type: List[RemovableHandle]
+    hooks = None if input_data is None else []  # type: Optional[List[RemovableHandle]]
     idx = {}  # type: Dict[int, int]
-    apply_hooks(model, model, depth, summary_list, hooks, idx, batch_dim)
+    apply_hooks(model, model, batch_dim, depth, summary_list, idx, hooks)
 
-    if device is None:
-        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
+    if input_data is not None:
+        if device is None:
+            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
 
+        x, input_size = process_input_data(input_data, batch_dim, device, dtypes)
+        args, kwargs = set_device(args, device), set_device(kwargs, device)
+        try:
+            with torch.no_grad():
+                _ = model.to(device)(*x, *args, **kwargs)  # type: ignore
+        except Exception:
+            executed_layers = [layer for layer in summary_list if layer.executed]
+            print("Failed to run torchsummary, executed layers up to: {}".format(executed_layers))
+            raise
+        finally:
+            if hooks is not None:
+                for hook in hooks:
+                    hook.remove()
+
+    formatting = FormattingOptions(branching, depth, verbose, col_names, col_width)
+    formatting.set_layer_name_width(summary_list)
+    results = ModelStatistics(summary_list, input_size, formatting)
+    if verbose > Verbosity.QUIET.value:
+        print(results)
+    return results
+
+
+def validate_user_params(
+    input_data: INPUT_DATA_TYPE, col_names: Sequence[str], verbose: int
+) -> None:
+    """ Raise exceptions if the user's input is invalid. """
+    if verbose not in (0, 1, 2):
+        raise ValueError("Verbose must be either 0 (quiet), 1 (default), or 2 (verbose).")
+
+    for col in col_names:
+        if col not in HEADER_TITLES.keys():
+            raise ValueError("Column {} is not a valid column name.".format(col))
+        if input_data is None and col not in ("num_params", "kernel_size"):
+            raise ValueError("You must pass input_data in order to use column {}".format(col))
+
+
+def set_device(data: Any, device: torch.device) -> Any:
+    """ Sets device for all input types and collections of input types. """
+    if torch.is_tensor(data):
+        return data.to(device, non_blocking=True)
+
+    # Recursively apply to collection items
+    elem_type = type(data)
+    if isinstance(data, Mapping):
+        return elem_type({k: set_device(v, device) for k, v in data.items()})
+    if isinstance(data, tuple) and hasattr(data, "_fields"):  # Named tuple
+        return elem_type(*(set_device(d, device) for d in data))
+    if isinstance(data, Sequence) and not isinstance(data, str):
+        return elem_type([set_device(d, device) for d in data])
+    # Data is neither a tensor nor a collection
+    return data
+
+
+def process_input_data(
+    input_data: INPUT_DATA_TYPE,
+    batch_dim: Optional[int],
+    device: torch.device,
+    dtypes: Optional[List[torch.dtype]],
+) -> Tuple[INPUT_DATA_TYPE, CORRECTED_INPUT_SIZE_TYPE]:
+    """ Create sample input data and the corrected input size. """
     if isinstance(input_data, torch.Tensor):
         input_size = get_correct_input_sizes(input_data.size())
         x = [input_data.to(device)]
@@ -102,7 +175,7 @@ def summary(
         if all(isinstance(data, torch.Tensor) for data in input_data):
             input_sizes = [data.size() for data in input_data]  # type: ignore
             input_size = get_correct_input_sizes(input_sizes)
-            x = [data.to(device) for data in input_data]  # type: ignore
+            x = set_device(input_data, device)
         else:
             if dtypes is None:
                 dtypes = [torch.float] * len(input_data)
@@ -117,34 +190,12 @@ def summary(
             "please submit a GitHub issue."
         )
 
-    args = tuple([t.to(device) if torch.is_tensor(t) else t for t in args])
-    kwargs = {k: kwargs[k].to(device) if torch.is_tensor(kwargs[k]) else k for k in kwargs}
-
-    try:
-        with torch.no_grad():
-            _ = model.to(device)(*x, *args, **kwargs)
-    except Exception:
-        print(
-            "Failed to run torchsummary, printing sizes of executed layers: {}".format(
-                [l for l in summary_list if l.executed]
-            )
-        )
-        raise
-    finally:
-        for hook in hooks:
-            hook.remove()
-
-    formatting = FormattingOptions(branching, depth, verbose, col_names, col_width)
-    formatting.set_layer_name_width(summary_list)
-    results = ModelStatistics(summary_list, input_size, formatting)
-    if verbose > Verbosity.QUIET.value:
-        print(results)
-    return results
+    return x, input_size
 
 
 def get_input_tensor(
     input_size: CORRECTED_INPUT_SIZE_TYPE,
-    batch_dim: int,
+    batch_dim: Optional[int],
     dtypes: List[torch.dtype],
     device: torch.device,
 ) -> List[torch.Tensor]:
@@ -152,20 +203,19 @@ def get_input_tensor(
     x = []
     for size, dtype in zip(input_size, dtypes):
         # add batch_size of 2 for BatchNorm
-        if isinstance(size, (list, tuple)):
-            # Case: input_tensor is a list of dimensions
-            input_tensor = torch.rand(*size)
+        input_tensor = torch.rand(*size)
+        if batch_dim is not None:
             input_tensor = input_tensor.unsqueeze(dim=batch_dim)
             input_tensor = torch.cat([input_tensor] * 2, dim=batch_dim)
-        result = input_tensor.to(device).type(dtype)
-        if isinstance(result, torch.Tensor):
-            x.append(result)
+        x.append(input_tensor.to(device).type(dtype))
     return x
 
 
 def get_correct_input_sizes(input_size: INPUT_SIZE_TYPE) -> CORRECTED_INPUT_SIZE_TYPE:
-    """ Convert input_size to the correct form, which is a list of tuples.
-    Also handles multiple inputs to the network. """
+    """
+    Convert input_size to the correct form, which is a list of tuples.
+    Also handles multiple inputs to the network.
+    """
 
     def flatten(nested_array: INPUT_SIZE_TYPE) -> Generator:
         """ Flattens a nested array. """
@@ -175,8 +225,8 @@ def get_correct_input_sizes(input_size: INPUT_SIZE_TYPE) -> CORRECTED_INPUT_SIZE
             else:
                 yield item
 
-    assert input_size
-    assert all(size > 0 for size in flatten(input_size)), "Negative size found in input_data."
+    if not input_size or any(size <= 0 for size in flatten(input_size)):
+        raise ValueError("Input_data is invalid, or negative size found in input_data.")
 
     if isinstance(input_size, list) and isinstance(input_size[0], int):
         return [tuple(input_size)]
@@ -190,17 +240,20 @@ def get_correct_input_sizes(input_size: INPUT_SIZE_TYPE) -> CORRECTED_INPUT_SIZE
 def apply_hooks(
     module: nn.Module,
     orig_model: nn.Module,
+    batch_dim: Optional[int],
     depth: int,
     summary_list: List[LayerInfo],
-    hooks: List[RemovableHandle],
     idx: Dict[int, int],
-    batch_dim: int,
+    hooks: Optional[List[RemovableHandle]],
     curr_depth: int = 0,
     parent_info: Optional[LayerInfo] = None,
 ) -> None:
-    """ Recursively adds hooks to all layers of the model. """
-    fallback_info = LayerInfo(module, curr_depth, None, parent_info)  # if layer is not called
-    info = fallback_info  # only define it for type checking
+    """
+    If input_data is provided, recursively adds hooks to all layers of the model.
+    Else, fills summary_list with layer info without computing a forward pass through the network.
+    """
+    # Fallback is used if the layer's hook is never called, in Module Lists, for example.
+    info = LayerInfo(module, curr_depth, None, parent_info)
 
     def pre_hook(module: nn.Module, inputs: Any) -> None:
         """ Create a LayerInfo object to aggregate information about that layer. """
@@ -208,33 +261,27 @@ def apply_hooks(
         nonlocal info
         idx[curr_depth] = idx.get(curr_depth, 0) + 1
         info = LayerInfo(module, curr_depth, idx[curr_depth], parent_info)
-        info.depth_index = idx[curr_depth]
         info.check_recursive(summary_list)
         summary_list.append(info)
 
     def hook(module: nn.Module, inputs: Any, outputs: Any) -> None:
         """ Update LayerInfo after forward pass. """
-        del module, inputs
-        info.calculate_output_size(outputs, batch_dim)
-        info.calculate_num_params()
+        del module
+        info.input_size = info.calculate_size(inputs, batch_dim)
+        info.output_size = info.calculate_size(outputs, batch_dim)
+        info.calculate_macs()
         info.executed = True
 
-    # ignore Sequential and ModuleList and other containers
     submodules = [m for m in module.modules() if m is not orig_model]
     if module != orig_model or isinstance(module, LAYER_MODULES) or not submodules:
-        hooks.append(module.register_forward_pre_hook(pre_hook))
-        hooks.append(module.register_forward_hook(hook))
+        if hooks is None:
+            pre_hook(module, None)
+        else:
+            hooks.append(module.register_forward_pre_hook(pre_hook))
+            hooks.append(module.register_forward_hook(hook))
 
     if curr_depth <= depth:
         for child in module.children():
             apply_hooks(
-                child,
-                orig_model,
-                depth,
-                summary_list,
-                hooks,
-                idx,
-                batch_dim,
-                curr_depth + 1,
-                fallback_info,
+                child, orig_model, batch_dim, depth, summary_list, idx, hooks, curr_depth + 1, info
             )
